"""Test s3 file poller."""

import os
from collections.abc import Generator
from datetime import datetime, timedelta, timezone

import pytest
from moto.moto_server.threaded_moto_server import ThreadedMotoServer
from s3fs import S3FileSystem
from upath import UPath

from pytroll_watchers.s3_poller import _poll_files

filenames = ["sat_20250828T1000.dat", "sat_20250828T1010.dat"]
file_pattern = "{platform_name}_{start_time:%Y%m%dT%H%M}.dat"


@pytest.fixture(scope="session", autouse=True)
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"  # noqa
    os.environ["AWS_SECURITY_TOKEN"] = "testing"  # noqa
    os.environ["AWS_SESSION_TOKEN"] = "testing"  # noqa
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"


# Note: pick an appropriate fixture "scope" for your use case
@pytest.fixture(scope="session")
def endpoint() -> Generator[str]:
    """Fixture to run a mocked AWS server for testing."""
    # Note: pass `port=0` to get a random free port.
    server = ThreadedMotoServer(port=0)
    server.start()
    host, port = server.get_host_and_port()
    yield f"http://{host}:{port}"
    server.stop()


@pytest.fixture
def bucket(endpoint: str) -> Generator[str]:
    """Create a bucket."""
    from botocore.session import Session

    session = Session()
    client = session.create_client("s3", endpoint_url=endpoint)
    bucketname = "test"
    client.create_bucket(Bucket=bucketname, ACL="public-read")
    yield bucketname
    client.delete_bucket(Bucket=bucketname)


def create_file(s3: S3FileSystem, filename: str):
    """Create a dummy file."""
    data = "very important data"
    with s3.open(filename, "w") as f:
        f.write(data)


@pytest.fixture
def some_files(endpoint: str, bucket: str):
    """Create a couple of files."""
    s3 = S3FileSystem(endpoint_url=endpoint)
    for filename in filenames:
        create_file(s3, bucket + "/" + filename)
    yield

    for filename in filenames:
        s3.rm_file(bucket + "/" + filename)


def test_file_poll(endpoint: str, bucket: str, some_files):
    """Test polling."""
    assert list(_poll_files(bucket, endpoint_url=endpoint).keys()) == [bucket + "/" + f for f in filenames]
    assert list(_poll_files(bucket, "sat_{time}",
                            endpoint_url=endpoint).keys()) == [bucket + "/" + f for f in filenames]
    assert list(_poll_files(bucket, filenames[0], endpoint_url=endpoint).keys()) == [bucket + "/" + filenames[0]]


def test_poll_files_since(endpoint: str, bucket: str, some_files):
    """Test polling since."""
    before = datetime.now(timezone.utc) - timedelta(seconds=2)
    assert list(_poll_files(bucket, start_date=before, endpoint_url=endpoint)) == [bucket + "/" + f for f in filenames]
    assert list(_poll_files(bucket, start_date=datetime.now(timezone.utc), endpoint_url=endpoint)) == []


def test_file_generator(endpoint: str, bucket: str):
    """Test the file generator."""
    from pytroll_watchers.s3_poller import file_generator
    s3 = S3FileSystem(endpoint_url=endpoint)
    gen = file_generator(bucket, start_from=timedelta(seconds=2),
                         polling_interval=timedelta(seconds=0.01),
                         storage_options=dict(endpoint_url=endpoint))
    filenames = [f"f{num}.file" for num in range(10)]
    try:
        for filename in filenames:
            create_file(s3, bucket + "/" + filename)
            f, _ = next(gen)
            assert f == UPath(bucket + "/" + filename, protocol="s3", endpoint_url=endpoint)
    finally:
        for filename in filenames:
            s3.rm(bucket + "/" + filename)


def test_file_generator_with_dicts(endpoint: str, bucket: str):
    """Test the file generator."""
    from pytroll_watchers.s3_poller import file_generator
    s3 = S3FileSystem(endpoint_url=endpoint)
    gen = file_generator(bucket, start_from=dict(seconds=2),
                         polling_interval=dict(seconds=0.01),
                         storage_options=dict(endpoint_url=endpoint))
    try:
        for filename in ["f1", "f2"]:
            create_file(s3, bucket + "/" + filename)
            f, _ = next(gen)
            assert f == UPath(bucket + "/" + filename, protocol="s3", endpoint_url=endpoint)
    finally:
        for filename in ["f1", "f2"]:
            s3.rm(bucket + "/" + filename)


def test_generate_download_links(endpoint: str, bucket: str, some_files):
    """Test link generation."""
    from pytroll_watchers.s3_poller import generate_download_links
    for (path, mda), f in zip(generate_download_links(bucket, file_pattern=file_pattern,
                                                      storage_options=dict(endpoint_url=endpoint)),
                              filenames,
                              strict=False):
        assert path.path == bucket + "/" + f
        assert path.protocol == "s3"
        assert path.storage_options == dict(endpoint_url=endpoint)
        assert mda["platform_name"] == "sat"
        assert mda["start_time"] in (datetime(2025, 8, 28, 10, 0), datetime(2025, 8, 28, 10, 10))


def test_generate_download_links_since(endpoint: str, bucket: str, some_files):
    """Test link generation from date."""
    from pytroll_watchers.s3_poller import generate_download_links_since
    assert list(generate_download_links_since(bucket, start_from=datetime.now(timezone.utc),
                                              storage_options=dict(endpoint_url=endpoint))) == []


def test_file_publisher(endpoint:str, bucket: str, some_files):
    """Test the file publisher."""
    from posttroll.testing import patched_publisher

    from pytroll_watchers.s3_poller import file_publisher

    config = dict(fs_config=dict(bucket_name=bucket,
                                 storage_options=dict(endpoint_url=endpoint),
                                 start_from=dict(seconds=2),
                                 polling_interval=timedelta(0)),
                  publisher_config=dict(name="s3"),
                  message_config=dict(subject="hej",
                                      atype="file"))
    with patched_publisher() as messages:
        file_publisher(config)
        assert filenames[0] in messages[0]
        assert filenames[1] in messages[1]



"""Functions and classes for performing message selection.

Selection in this context means making sure only one message refering to some file will be published further.

This is useful when multiple source for the same data are sending messages (eg two reception servers for eumetcast) but
only one of each file is needed for further processing.

To check if two messages refer to the same data, the *uid* metadata of the messages is used.

A command-line script is also made available by this module. It is called ``pytroll-selector``::

    usage: pytroll-selector [-h] [-l LOG_CONFIG] config

    Selects unique messages (based on uid) from multiple sources.

    positional arguments:
      config                The yaml config file.

    options:
      -h, --help            show this help message and exit
      -l LOG_CONFIG, --log-config LOG_CONFIG
                            The yaml config file for logging.

    Thanks for using pytroll-selector!

An example config file to use with this script is the following::

    selector_config:
      ttl: 30
    publisher_config:
      name: hrit_selector
    subscriber_config:
      addresses:
        - tcp://eumetcast_reception_1:9999
        - tcp://eumetcast_reception_2:9999
      nameserver: false
      topics:
        - /1b/hrit-segment/0deg

The different sections are passed straight on to :py:func:`run_selector`, so check it to have more information about
what to pass to it.


"""

import argparse
import logging
from contextlib import closing
from threading import Timer

import yaml
from posttroll.publisher import create_publisher_from_dict_config
from posttroll.subscriber import create_subscriber_from_dict_config

from pytroll_watchers.main_interface import configure_logging

logger = logging.getLogger(__name__)


class TTLDict:
    """A simple dictionary-like object that discards items older than a time-to-live.

    Not thread-safe.

    Args:
        ttl: the time to live of the stored items in seconds.
    """

    def __init__(self, ttl=300):
        """Set up the instance.

        Args:
            ttl: the time to live of the stored items in seconds.
        """
        self._data = dict()
        self._ttl = ttl

    def __getitem__(self, key):
        """Get the value corresponding to *key*."""
        return self._data[key]

    def __setitem__(self, key, value):
        """Set the *value* corresponding to *key*."""
        if key not in self._data:
            self._data[key] = value
            Timer(self._ttl, self._data.pop, (key, None)).start()

    def __contains__(self, key):
        """Check if key is already present."""
        try:
            _ = self[key]
            return True
        except KeyError:
            return False


def running_selector(selector_config, subscriber_config):
    """Generate selected messages.

    The aim of this generator is to skip messages that are duplicates to already processed messages.
    Duplicate in this context means messages referring to the same file (even if stored in different locations).

    Args:
        selector_config: a dictionary providing a ttl in seconds, otherwise it defaults to 300 seconds (5 minutes).
        subscriber_config: a dictionary of arguments to pass to
          :py:func:`~posttroll.subscriber.create_subscriber_from_dict_config`.

    Yields:
        JSON representations of posttroll messages.
    """
    ttl_dict = TTLDict(**selector_config)

    for msg in _data_messages(subscriber_config):
        key = unique_key(msg)
        msg_string = str(msg)

        if key not in ttl_dict:
            ttl_dict[key] = msg_string
            logger.info(f"New content {msg_string}")
            yield msg_string
        else:
            logger.debug(f"Discarded {msg_string}")


def _data_messages(subscriber_config):
    """Generate messages referring to new data from subscriber settings."""
    subscriber = create_subscriber_from_dict_config(subscriber_config)

    with closing(subscriber):
        for msg in subscriber.recv():
            if msg.type != "file":
                continue
            yield msg


def unique_key(msg):
    """Identify the content of the message with a unique key."""
    return msg.data["uid"]


def _run_selector_with_managed_dict_server(selector_config, subscriber_config, publisher_config):
    """Run the selector with a managed ttldict server."""
    publisher = create_publisher_from_dict_config(publisher_config)
    publisher.start()
    with closing(publisher):
        for msg in running_selector(selector_config, subscriber_config):
            publisher.send(msg)


def run_selector(selector_config, subscriber_config, publisher_config):
    """Run the selector.

    The aim of the selector is to skip messages that are duplicates to already published messages.
    Duplicate in this context means messages referring to the same file (even if stored in different locations).

    Messages that refer to new files will be published.

    Args:
        selector_config: A dictionary providing a *ttl* for the
          selector as seconds, so that incoming messages are forgotten after that time.
          If not provided, the ttl defaults to 300 seconds (5 minutes).
        subscriber_config: a dictionary of arguments to pass to
          :py:func:`~posttroll.subscriber.create_subscriber_from_dict_config`. The subscribtion is used as a source for
          messages to process.
        publisher_config: a dictionary of arguments to pass to
          :py:func:`~posttroll.publisher.create_publisher_from_dict_config`. This publisher will send the selected
          messages.

    """
    _run_selector_with_managed_dict_server(selector_config, subscriber_config, publisher_config)


def cli(args=None):
    """Command line interface."""
    parser = argparse.ArgumentParser(
                    prog="pytroll-selector",
                    description="Selects unique messages (based on uid) from multiple sources.",
                    epilog="Thanks for using pytroll-selector!")

    parser.add_argument("config", type=str, help="The yaml config file.")
    parser.add_argument("-l", "--log-config", type=str, help="The yaml config file for logging.", default=None)

    parsed = parser.parse_args(args)


    log_config_filename = parsed.log_config
    configure_logging(log_config_filename)

    config_file = parsed.config

    with open(config_file) as fd:
        config_dict = yaml.safe_load(fd.read())

    selector_config = config_dict.get("selector_config", {})
    subscriber_config = config_dict.get("subscriber_config", {})
    publisher_config = config_dict.get("publisher_config", {})

    return run_selector(selector_config, subscriber_config, publisher_config)



"""Poller for S3 object stores.

Configuration example:

.. code-block:: yaml

    backend: s3
    fs_config:
      bucket_name: sat/L1B
      file_pattern: "SAT_{platform_name}-{start_time:%Y%m%d%H%M%S}_{end_time:%Y%m%d%H%M%S}.nc"
      storage_options:
        profile: sat-store
      polling_interval:
        seconds: 10
      start_from:
        days: 1
    publisher_config:
      name: sat_watcher
      nameservers: false
      port: 3000
    message_config:
      subject: /segment/s1/l1b/
      atype: file
    data_config:
      include_dir_in_uid: true

in this example, the credentials and endpoint are provided through a profile that needs to be defined in the .aws files,
for example, in `.aws/credentials`::

    ...
    [sat-store]
    aws_access_key_id=...
    aws_secret_access_key=...

and in `.aws/config`::

    ...
    [profile sat-store]
    services = sat-store-s3

    [services sat-store-s3]
    S3 =
      endpoint_url = https://sat.store.org


"""
import logging
import os
from collections import deque
from collections.abc import Generator
from datetime import datetime, timedelta, timezone
from typing import Any

import s3fs
from trollsift import Parser
from upath import UPath

from pytroll_watchers.common import run_every
from pytroll_watchers.publisher import file_publisher_from_generator

logger = logging.getLogger(__name__)



def file_publisher(config: dict[str, Any]):
    """Publish files coming from local filesystem events.

    Args:
        config: the configuration dictionary, containing in particular an fs_config section, which is the configuration
        for the filesystem watching, will be passed as argument to `file_generator`. The other sections are passed
        further to ``file_publisher_from_generator``.
    """
    fs_config = config["fs_config"]
    logger.info(f"Starting polling on s3 for '{fs_config['bucket_name']}'")
    generator = file_generator(**fs_config)
    return file_publisher_from_generator(generator, config)


def file_generator(bucket_name: str,
                   polling_interval: timedelta | dict[str, float],
                   file_pattern: str | None = None,
                   start_from: timedelta | dict[str, float] | None = None,
                   storage_options: None | dict[str, bool | int | str]=None) -> Generator[tuple[UPath, dict[str, Any]]]:
    """Generate file UPaths and metadata for a given bucket by polling."""
    if isinstance(polling_interval, dict):
        polling_interval = timedelta(**polling_interval)

    if isinstance(start_from, dict):
        start_from_delta = timedelta(**start_from)
    else:
        start_from_delta: timedelta = start_from or timedelta(0)

    filename_cache: deque[UPath] = deque(maxlen=100)

    last_pub_date: datetime = datetime.now(timezone.utc) - start_from_delta
    for next_check in run_every(polling_interval):
        new_pub_date =  None
        for f, mda in generate_download_links_since(bucket_name, file_pattern, last_pub_date, storage_options):
            if f in filename_cache:
                continue
            new_pub_date = mda["LastModified"]
            filename_cache.append(f)
            yield f, mda
        logger.info("Finished polling.")
        if next_check > datetime.now(timezone.utc):
            logger.info(f"Next iteration at {next_check}")
        last_pub_date = new_pub_date or last_pub_date


def generate_download_links(bucket_name: str,
                            file_pattern: str | None = None,
                            storage_options: dict[str, Any] | None = None) -> Generator[tuple[UPath, dict[str, Any]]]:
    """Generate download links."""
    return generate_download_links_since(bucket_name, file_pattern, None, storage_options)

def generate_download_links_since(bucket_name: str,
                                  file_pattern: str | None = None,
                                  start_from: datetime | None = None,
                                  storage_options: dict[str, Any] | None = None):
    """Generate download links since date."""
    for f, mda in sorted(_poll_files(bucket_name, file_pattern, start_from,**(storage_options or dict())).items(),
                         key=(lambda x: x[1]["LastModified"])):
        s3path = UPath(f, protocol="s3", **(storage_options or dict()))
        yield s3path, mda

def _poll_files(bucket_name: str,
               file_pattern: str | None = None,
               start_date: datetime | None = None,
               **storage_options: Any,
               ) -> dict[str, dict[str, str | datetime | list[str] | int]]:
    """Poll files from s3."""
    s3 = s3fs.S3FileSystem(skip_instance_cache=True, **storage_options)
    if file_pattern:
        parser = Parser(file_pattern)
        file_pattern = parser.globify()
    flist = s3.glob(bucket_name + "/" + (file_pattern or "*"), detail=True)
    if start_date:
        flist = {f: mda for f, mda in flist.items() if mda["LastModified"] >= start_date}
    if file_pattern:
        for f, mda in flist.items():
            mda.update(parser.parse(os.path.basename(f)))
    return flist
